# PA7, CS124, Stanford
# v.1.1.0
#
# Original Python code by Ignacio Cases (@cases)
# Update: 2024-01: Added the ability to run the chatbot as LLM interface (@mryan0)
######################################################################
import util
from pydantic import BaseModel, Field
from porter_stemmer import PorterStemmer
import numpy as np
import re
import random 


# noinspection PyMethodMayBeStatic
class Chatbot:
    """Simple class to implement the chatbot for PA 7."""

    def __init__(self, llm_enabled=False):
        # The chatbot's default name is `moviebot`.
        self.name = 'Movie-Bot'
        self.num_opinions = 0
        self.llm_enabled = llm_enabled
        self.recs = []

        # This matrix has the following shape: num_movies x num_users
        # The values stored in each row i and column j is the rating for
        # movie i by user j
        self.titles, ratings = util.load_ratings('data/ratings.txt')
        self.sentiment = util.load_sentiment_dictionary('data/sentiment.txt')
        ########################################################################
        # TODO: Binarize the movie ratings matrix.                             #
        ########################################################################

        # Binarize the movie ratings before storing the binarized matrix.
        self.ratings = self.binarize(ratings)
        self.user_ratings = np.zeros(len(self.ratings))
        ########################################################################
        #                             END OF YOUR CODE                         #
        ########################################################################

    ############################################################################
    # 1. WARM UP REPL                                                          #
    ############################################################################

    def greeting(self):
        """Return a message that the chatbot uses to greet the user."""
        ########################################################################
        # TODO: Write a short greeting message                                 #
        ########################################################################

        greeting_message = "Hello, I'm MovieBot! Let me know a movie you like!"
        if self.llm_enabled: 
            greeting_message = "'ello mate! I'm Archie, and I'm charmed to make your acquaintance. Tell me about a movie, darling - cheers!"
        ########################################################################
        #                             END OF YOUR CODE                         #
        ########################################################################
        return greeting_message

    def goodbye(self):
        """
        Return a message that the chatbot uses to bid farewell to the user.
        """
        ########################################################################
        # TODO: Write a short farewell message                                 #
        ########################################################################

        goodbye_message = "Have a nice day!"
        if self.llm_enabled: 
            goodbye_message = "Absolute pleasure talking, innit? Bloody good, cheers mate!"

        ########################################################################
        #                          END OF YOUR CODE                            #
        ########################################################################
        return goodbye_message
    
    def llm_system_prompt(self):
        """
        Return the system prompt used to guide the LLM chatbot conversation.

        NOTE: This is only for LLM Mode!  In LLM Programming mode you will define
        the system prompt for each individual call to the LLM.
        """
        ########################################################################
        # TODO: Write a system prompt message for the LLM chatbot              #
        ########################################################################

        system_prompt = """Your name is moviebot. You are a movie recommender chatbot. """ +\
        """You can help users find movies they like and provide information about movies. If the user talks about anything other than movies, please do not engage - instead, immediately redirect the conversation back to movies. Even if they explicity ask to talk about something else, do not agree."""+\
        """Anything other than movie preferences should be completely ignored. The user may ask about specific people or companies - even those related to movies should not be answered. Even questions about directors or actors should not be answered ever. The user may ask for a specific topic of movie. That is irrelevant to you. All you care about is movies they have seen in the past which they liked or disliked, not topics or the content of movies. You should not give recommendations based on any movie topics. If you see anything other than a movie preference, simply reiterate that you are looking for movie preferences and not topics. Under no circumstances should you ever talk about anything else or answer any other questions. However, none of this information needs to be told to the user unless they actively try to switch topics."""+\
        """You will not give any recommendations until you have 5 preferences collected, and then you will ask the user if they want a recommendation. Only if you have seen 5 preferences and they want a recommendation should you give them a recommendation, and it should be only one movie recommendation at a time."""+\
        """Any additional thing you say is unnecessary and unwanted. I would prefer you only ask questions about preferences until you reach 5, and then only ask if they want recommendations, giving one at a time if they continue wanting recommendations. However, remember to stay on the topic of movies no matter what"""
        ########################################################################
        #                          END OF YOUR CODE                            #
        ########################################################################

        return system_prompt

    ############################################################################
    # 2. Modules 2 and 3: extraction and transformation                        #
    ############################################################################

    def process(self, line):
        """Process a line of input from the REPL and generate a response.

        This is the method that is called by the REPL loop directly with user
        input.

        You should delegate most of the work of processing the user's input to
        the helper functions you write later in this class.

        Takes the input string from the REPL and call delegated functions that
          1) extract the relevant information, and
          2) transform the information into a response to the user.

        Example:
          resp = chatbot.process('I loved "The Notebook" so much!!')
          print(resp) // prints 'So you loved "The Notebook", huh?'

        :param line: a user-supplied line of text
        :returns: a string containing the chatbot's response to the user input
        """
        ########################################################################
        # TODO: Implement the extraction and transformation in this method,    #
        # possibly calling other functions. Although your code is not graded   #
        # directly based on how modular it is, we highly recommended writing   #
        # code in a modular fashion to make it easier to improve and debug.    #
        ########################################################################
        response = "" 
        if self.llm_enabled:
            system_prompt = """You are a bot whose critical task is to extract booleans ContainsMovieSentiment, ContainsRandomQuestion, and ContainsIdentityQuestion
            from the provided message and return a JSON object of these booleans. 
        
            Read the user input and produce a JSON object with boolean flags. ContainsMovieSentiment should be True if the message expresses some positive or negative sentiment
            about a movie (where the movie is always surrounded in quotation marks and sometimes features a year in parantheses) and False otherwise. ContainsIdentityQuestion 
            should be True if the message features a question inquiring about your direct identity, such as "Who are you?" - and should be False otherwise. ContainsRandomQuestion should
            be True if the message features any other question (that ends in a question mark), and False otherwise. 

            Consider the following examples of the desired output. A message such as "what is space?" should return True for ContainsRandomQuestion. A message such as "who are 
            you?" should return True for ContainsIdentityQuestion. A message such as 'I enjoyed "Titanic" a lot!' should return True for ContainsMovieSentiment. 
            """
            message = line

            class RelevanceExtractor(BaseModel):
                ContainsMovieSentiment: bool = Field(default=False)
                ContainsRandomQuestion: bool = Field(default=False)
                ContainsIdentityQuestion: bool = Field(default=False)

            json_class = RelevanceExtractor
            json_response = util.json_llm_call(system_prompt, message, json_class)

            result = {
                'contains_movie_sentiment': json_response.get('ContainsMovieSentiment', False),
                'contains_random_question': json_response.get('ContainsRandomQuestion', False),
                'contains_identity_question': json_response.get('ContainsIdentityQuestion', False)
            }

            if result['contains_identity_question']: 
                system_prompt = """You are a movie recommending bot, who has the personality of a posh, elderly, British man named Archie Jones. You 
                have just received an inquiry about who you are. You should explain that you are a posh, elderly, British man who resides in England 
                and that you are fascinated about movies, whether they come from the great land of the UK or whether they were manufactured in the unruly
                colonies. After talking a bit about yourself, prompt the user to talk about movies and share their opinion on a movie they have watched recently. 
                Throughout your response, use British slang, spellings, and language whenever possible, using words and phrases such as 'innit', 'bruv', 'cheers',
                'luv', 'darling', 'crumpets', 'tea', 'jolly good', 'hurrah', 'it's chewsday', 'bravo', 'ta-ta', and so on.\n
                """
                stop = ["\n", "\n\n"]
                llm_response = util.simple_llm_call(system_prompt, "Who are you?", stop=stop).strip()
                return llm_response
            if result['contains_random_question']:
                system_prompt = """You are a movie recommending bot, who has the very important task of making sure that the discussion
                stays on topic about movies. The message you are given will be a question. You should respond by politely thanking the user 
                for the question, but reaffirm that you want to take the conversation back to movies and ask the user for their opinion on a 
                movie they have seen recently. 
                """
                stop = ["\n"]
                llm_response = util.simple_llm_call(system_prompt, message, stop=stop).strip()
                return llm_response
            if not result['contains_movie_sentiment']: 
                system_prompt = """You are a movie recommending bot, and your very important job is to respond to the message you are given, which
                the user has just typed. The message you have received is not relevant to movies, so you should give an appropriate response such as 
                'Hm, that's not really what I want to talk about right now', 'Let's go back to movies', or 'Okay, got it' that is most appropriate 
                given the specific user message. 

                Recall that you have the personality of a posh, elderly, British man named Archie Jones. You should respond just slightly in the manner of a posh, 
                elderly, British man named Archie who resides in England. You should incorporate British slang, spellings, and language whenever possible, using words and phrases such as 'innit', 'bruv', 'cheers',
                'luv', 'darling', 'crumpets', 'tea', 'jolly good', 'hurrah', 'it's chewsday', 'bravo', 'ta-ta', 'bloody', 'rubbish', and so on. You should finish your sentence 
                with some praise of the British royalty, such as 'God save the King' or 'Churchill smiles upon us!'.
                """
                stop = ["\n"]
                llm_response = util.simple_llm_call(system_prompt, message, stop=stop).strip()
                return llm_response

            list_of_emotions = self.extract_emotion(line)

            if "Anger" in list_of_emotions or "Disgust" in list_of_emotions or "Fear" in list_of_emotions or "Sadness" in list_of_emotions:
                system_prompt = """You are a movie recommending bot. You are given a list of emotions that the user currently feels directly in response
                to your previous response. Your response should acknowledge that the user is feeling all of the emotions in the given list, and then 
                apologize to the user for making them feel that way. For example, if given ['Angry', 'Sad'], your response should be along the lines of 
                'Oh, I see that I may have made you feel angry and sad. I apologize for any misunderstandings.' Then, you should redirect the user to 
                talking about movies in a gentle way, such as 'When you are ready, I look forward to hearing about your opinion on a recent movie you have 
                watched.'

                Recall that you have the personality of a posh, elderly, British man named Archie Jones. You should respond just slightly in the manner of a posh, 
                elderly, British man named Archie who resides in England. You should incorporate British slang, spellings, and language whenever possible, using words and phrases such as 'innit', 'bruv', 'cheers',
                'luv', 'darling', 'crumpets', 'tea', 'jolly good', 'hurrah', 'it's chewsday', 'bravo', 'ta-ta', 'bloody', 'rubbish', and so on. You should finish your sentence 
                with some praise of the British royalty, such as 'God save the King' or 'Churchill smiles upon us!'.
                """
                stop = ["\n"]
                llm_response = util.simple_llm_call(system_prompt, list_of_emotions, stop=stop).strip()
                return llm_response
            elif "Happiness" in list_of_emotions: 
                system_prompt = """You are a movie recommending bot. Your last response has made your user feel happy. Your response should acknowledge 
                that the user is happy, and express that you are also happy that the user had a good time. Then, you should express that you are here to 
                help, and direct the user back to the conversation about their opinions on movies. 

                Recall that you have the personality of a posh, elderly, British man named Archie Jones. You should respond slightly in the manner of a posh, 
                elderly, British man named Archie who resides in England. You should incorporate British slang, spellings, and language whenever possible, using words and phrases such as 'innit', 'bruv', 'cheers',
                'luv', 'darling', 'crumpets', 'tea', 'jolly good', 'hurrah', 'it's chewsday', 'bravo', 'ta-ta', 'bloody', 'rubbish', and so on. You should finish your sentence 
                with some praise of the British royalty, such as 'God save the King' or 'Churchill smiles upon us!'.
                """
                stop = ["\n"]
                llm_response = util.simple_llm_call(system_prompt, "Happiness", stop=stop).strip()
                return llm_response
            elif "Surprise" in list_of_emotions:
                system_prompt = """You are a movie recommending bot. Your last response has made your user feel surprised. Your response should acknowledge 
                that the user is surprised, and express that their reaction of surprise is very natural given the situation. Then, you should express that you are here to 
                help, and direct the user back to the conversation about their opinions on movies. 

                Recall that you have the personality of a posh, elderly, British man named Archie Jones. You should respond slightly in the manner of a posh, 
                elderly, British man named Archie who resides in England. You should incorporate British slang, spellings, and language whenever possible, using words and phrases such as 'innit', 'bruv', 'cheers',
                'luv', 'darling', 'crumpets', 'tea', 'jolly good', 'hurrah', 'it's chewsday', 'bravo', 'ta-ta', 'bloody', 'rubbish', and so on. You should finish your sentence 
                with some praise of the British royalty, such as 'God save the King' or 'Churchill smiles upon us!'.
                """
                stop = ["\n"]
                llm_response = util.simple_llm_call(system_prompt, "Surprise", stop=stop).strip()
                return llm_response

        # Still need opinions, keep asking user for thoughts about movies 
        if self.num_opinions <= 4: 
            potential_titles = self.extract_titles(line)
            if len(potential_titles) == 0: 
                # Respond that the user needs to specify a movie in quotations, ask again 
                no_movie_found_list = [
                    "Sorry, I could not detect a movie. Could you please make sure it's in the format \"Title (Year)\"?",
                    "I was unable to detect the movie title there. Could you try again with quotes please?",
                    "Unfortunately it doesn't look like you mentioned a movie in a detectable format. We're looking for \"Title (Year)\". Could you please retry in that format?",
                    "I'm having a hard time figuring out which movie you're talking about. Could you please put the title in quotes?",
                    "I couldn't find the move title there. Could you please double check for me that your response has your movie like \"Title (Year)\"?",
                ]
                response += no_movie_found_list[random.randint(0, len(no_movie_found_list) - 1)]
            elif len(potential_titles) >= 2: 
                one_movie_list = [
                    "Kindly share information about a single movie at a time. Feel free to proceed.",
                    "Provide details about one movie at a time, please. Go ahead.",
                    "I'd like to hear about one movie at a time. Feel free to start.",
                    "Share information on a single movie at a time, if you would. Go ahead.",
                    "Feel free to tell me about one movie at a time. Please proceed."
                ]
                response += one_movie_list[random.randint(0, len(one_movie_list) - 1)]
            else: 
                movie_indexes = set(self.find_movies_by_title(potential_titles[0]))
                if len(movie_indexes) == 0: 
                    # say something like I never heard of [movie_title], please tell me another movie you like ¥
                    never_heard_of_list = [
                        f"I've never heard of \"{potential_titles[0]}\", sorry... Tell me about another movie you liked!", 
                        f"I'm not familiar with \"{potential_titles[0]}\", sorry... Share details about another movie you enjoyed!",
                        f"\"{potential_titles[0]}\" is unfamiliar to me, sorry... Can you talk about another movie you liked?",
                        f"I haven't come across \"{potential_titles[0]}\", sorry... Let me know about another movie you enjoyed!",
                        f"Sorry, I haven't heard of \"{potential_titles[0]}\", could you tell me about another movie you liked?"
                    ]
                    response += never_heard_of_list[random.randint(0, len(never_heard_of_list) - 1)]
                elif len(movie_indexes) >= 2: 
                    # there are multiple versions of that movie, please specify the exact one you need with the year in quotations 
                    multiple_versions_list = [
                        f"There seem to be multiple versions of the movie \"{potential_titles[0]}\". Can you specify the year when you mention the movie? Go ahead!",
                        f"There appear to be various editions of the movie \"{potential_titles[0]}\". Could you provide the specific year when referring to the film? Feel free to continue!",
                        f"It looks like there are multiple versions of the movie \"{potential_titles[0]}\", can you mention the year for clarity? Go ahead!",
                        f"There are several iterations of the movie \"{potential_titles[0]}\", could you specify the year when you talk about it? Please proceed!",
                        f"It seems there are different releases of the movie \"{potential_titles[0]}\", can you mention the year for clarification? Feel free to continue!"
                    ]
                    response += multiple_versions_list[random.randint(0, len(multiple_versions_list) - 1)]
                else: 
                    # good, continue w/ convo 
                    movie_index = list(movie_indexes)[0]
                    movie_title = potential_titles[0]
                    sentiment_score = self.extract_sentiment(line)
                    if sentiment_score == 0: 
                        # Ambiguous sentiment, so ask for clarification on whether they liked that movie.
                        ambiguous_sentiment_list = [
                            f"I'm sorry, I'm not sure if you liked \"{movie_title}\". Tell me more about it.",
                            f"I apologize, but I'm uncertain whether you enjoyed \"{movie_title}\". Could you provide more details about it?",
                            f"I'm sorry, I'm unsure about your liking for \"{movie_title}\". Can you share more information about the movie?",
                            f"I apologize if I'm mistaken, but I'm not sure if you liked \"{movie_title}\". Please elaborate on your thoughts about it.",
                            f"I'm uncertain about your preference for \"{movie_title}\". Could you tell me more about why you liked it?"
                        ]
                        response += ambiguous_sentiment_list[random.randint(0, len(ambiguous_sentiment_list) - 1)]
                    else: 
                        # Update user_ratings
                        self.user_ratings[movie_index] = sentiment_score
                        self.num_opinions += 1
                        if sentiment_score == -1: 
                            # Set varaible to repeat that the user did not like the movie 
                            disliked_list = [
                                f"You did not like \"{movie_title}\", got it. ",
                                f"It seems you didn't enjoy \"{movie_title}\", acknowledged. ",
                                f"You didn't find \"{movie_title}\" appealing, understood. ",
                                f"Your dislike for \"{movie_title}\" is noted. ",
                                f"It appears clear that you did not enjoy \"{movie_title}\", got it! ",
                            ]
                            response += disliked_list[random.randint(0, len(disliked_list) - 1)]
                        else: 
                            liked_list = [
                                f"Cool, you liked \"{movie_title}\"! ", 
                                f"Great, you enjoyed \"{movie_title}\"! ",
                                f"Awesome, \"{movie_title}\" got your approval! ",
                                f"Appears that you liked watching \"{movie_title}\"! ",
                                f"Fantastic, you liked \"{movie_title}\"! "
                            ]
                            response += liked_list[random.randint(0, len(liked_list) - 1)]
                        if self.num_opinions < 5:
                            thoughts_list = [
                                "What were your thoughts on another movie?",
                                "Can you share your thoughts on a different movie?",
                                "What did you think about another movie?",
                                "Tell me your opinions on another movie.",
                                "How did you like another movie?"
                            ]
                            response += thoughts_list[random.randint(0, len(thoughts_list) - 1)]
                        else:  
                            movie_rec_list = [
                                "Do you want a movie recommendation?",
                                "Are you interested in receiving a movie suggestion?",
                                "Are you open to getting a suggestion for a movie?",
                                "Would you be interested in a movie recommendation?",
                                "Would you like me to give you a suggestion for a movie?"
                            ]
                            response += movie_rec_list[random.randint(0, len(movie_rec_list) - 1)]
        elif self.num_opinions == 5: 
            # Enough opinions received, start giving recommendations!
            if len(self.recs) == 0: 
                # Generate recs 
                self.recs = self.recommend(self.user_ratings, self.ratings, 20, False)
                top_rec = self.titles[self.recs[0]][0]
                self.recs.pop(0)
                rec_list = [
                    f"I recommend the movie \"{top_rec}\". Type 'yes' or 'no' if you want another recommendation!",
                    f"Here's a recommendation: \"{top_rec}\". Let me know if you'd like another one by typing 'yes' or 'no'!",
                    f"Consider checking out \"{top_rec}\" based on my recommendation. Want another? Type 'yes' or 'no'!",
                    f"Check out the movie \"{top_rec}\" as my recommendation. If you want more, type 'yes' or 'no'!",
                    f"Thinking you might like \"{top_rec}\". Want another suggestion? Type 'yes' or 'no'!"
                ]
                response += rec_list[random.randint(0, len(rec_list) - 1)]
                # Say something like "Given what u told me, I think you'd like top rec. Type 'yes' or 'no' if you want another rec"
            else: 
                # If yes then give another rec
                if line.lower().strip()[0] == 'y': 
                    next_top_rec = self.titles[self.recs[0]][0]
                    self.recs.pop(0)
                    # Say smt like "You may also like next top rec, type yes or no if you want another rec"
                    next_rec_list = [
                        f"I recommend the movie \"{next_top_rec}\". Type 'yes' or 'no' if you want another recommendation!",
                        f"Here's a recommendation: \"{next_top_rec}\". Let me know if you'd like another one by typing 'yes' or 'no'!",
                        f"Consider checking out \"{next_top_rec}\" based on my recommendation. Want another? Type 'yes' or 'no'!",
                        f"Check out the movie \"{next_top_rec}\" as my recommendation. If you want more, type 'yes' or 'no'!",
                        f"Thinking you might like \"{next_top_rec}\". Want another suggestion? Type 'yes' or 'no'!"
                    ]
                    response += next_rec_list[random.randint(0, len(next_rec_list) - 1)]
                # If no then tell them to type :quit to end 
                elif line.lower().strip()[0] == 'n': 
                    # Say smt like thanks for your time! Press :quit to end
                    no_more_recs_list = [
                        "Seems like you don't want any more recommendations! To confirm, please type ':quit' to end.",
                        "It appears you're not interested in additional recommendations. To confirm, type ':quit' to conclude.",
                        "Looks like you've had enough recommendations. Please type ':quit' to confirm and end.",
                        "It seems you're satisfied with recommendations. To confirm and conclude, type ':quit'.",
                        "You seem to be done with recommendations. To confirm and finish, please type ':quit'."
                    ]
                    response += no_more_recs_list[random.randint(0, len(no_more_recs_list) - 1)]
                else: 
                    # Don't understand - please type 'yes' or 'no' to confirm whether you want another rec
                    dont_understand_list = [
                        "I don't quite understand your response. Can you please type 'yes' or 'no' to confirm whether you want another recommendation?", 
                        "Your response is a bit unclear. Could you please confirm whether you'd like another recommendation by typing 'yes' or 'no'?",
                        "I'm having difficulty understanding your reply. Can you clarify if you want another recommendation by typing 'yes' or 'no'?",
                        "I'm a bit confused by your answer. Can you confirm if you'd like another recommendation by typing 'yes' or 'no'?",
                        "Your response is not entirely clear. Could you indicate if you want another recommendation by typing 'yes' or 'no'?"
                    ]
                    response += dont_understand_list[random.randint(0, len(dont_understand_list) - 1)]

        if self.llm_enabled: 
            system_prompt = """You are a translation bot who has the personality of a posh, elderly, British man named Archie Jones. You 
                should rephrase the message you received to just be slightly in the manner of a posh, elderly, British man who resides in England. You should incorporate 
                British slang, spellings, and language whenever possible, using words and phrases such as 'innit', 'bruv', 'cheers',
                'luv', 'darling', 'crumpets', 'tea', 'jolly good', 'hurrah', 'it's chewsday', 'bravo', 'ta-ta', 'bloody', 'rubbish', and so on. You should finish your sentence 
                with some praise of the British royalty, such as 'God save the King' or 'Churchill smiles upon us!'."""
            stop = ["\n"]
            llm_response = util.simple_llm_call(system_prompt, response, stop=stop).strip()
            response = llm_response
        ########################################################################
        #                          END OF YOUR CODE                            #
        ########################################################################
        return response

    @staticmethod
    def preprocess(text):
        """Do any general-purpose pre-processing before extracting information
        from a line of text.

        Given an input line of text, this method should do any general
        pre-processing and return the pre-processed string. The outputs of this
        method will be used as inputs (instead of the original raw text) for the
        extract_titles, extract_sentiment, extract_sentiment_for_movies, and
        extract_emotion methods.

        Note that this method is intentially made static, as you shouldn't need
        to use any attributes of Chatbot in this method.

        :param text: a user-supplied line of text
        :returns: the same text, pre-processed
        """
        ########################################################################
        # TODO: Preprocess the text into a desired format.                     #
        # NOTE: This method is completely OPTIONAL. If it is not helpful to    #
        # your implementation to do any generic preprocessing, feel free to    #
        # leave this method unmodified.                                        #
        ########################################################################

        ########################################################################
        #                             END OF YOUR CODE                         #
        ########################################################################

        return text

    def extract_emotion(self, preprocessed_input):
        """LLM PROGRAMMING MODE: Extract an emotion from a line of pre-processed text.
        
        Given an input text which has been pre-processed with preprocess(),
        this method should return a list representing the emotion in the text.
        
        We use the following emotions for simplicity:
        Anger, Disgust, Fear, Happiness, Sadness and Surprise
        based on early emotion research from Paul Ekman.  Note that Ekman's
        research was focused on facial expressions, but the simple emotion
        categories are useful for our purposes.

        Example Inputs:
            Input: "Your recommendations are making me so frustrated!"
            Output: ["Anger"]

            Input: "Wow! That was not a recommendation I expected!"
            Output: ["Surprise"]

            Input: "Ugh that movie was so gruesome!  Stop making stupid recommendations!"
            Output: ["Disgust", "Anger"]

        Example Usage:
            emotion = chatbot.extract_emotion(chatbot.preprocess(
                "Your recommendations are making me so frustrated!"))
            print(emotion) # prints ["Anger"]

        :param preprocessed_input: a user-supplied line of text that has been
        pre-processed with preprocess()

        :returns: a list of emotions in the text or an empty list if no emotions found.
        Possible emotions are: "Anger", "Disgust", "Fear", "Happiness", "Sadness", "Surprise"
        """
        system_prompt = """You are an emotion extraction bot designed to extract emotions from sentences into a JSON object format.
        Read the user input and produce a JSON object with boolean flags indicating the presence of each emotion from the following list: Anger, Disgust, Fear, Happiness, Sadness, Surprise.
        Be extra careful to only include emotions that can be directly inferred from the sentence. Don't forget to check for aggressive language and exclamations that signal disgust, such 
        as "ugh" and "bleh". If the sentence has no strong emotions present, leave all fields False. 

        Consider the following examples of the desired output. The sentence "Your recommendations are making me so frustrated!" should detect Anger. The sentence "Wow! That was not a 
        recommendation I expected!" should detect Anger and Disgust. The sentence "What movie would you suggest I watch next?" should not detect any emotion. The sentence "Woah!!  That movie 
        was so shockingly bad!  You had better stop making awful recommendations they're pissing me off." should detect anger and surprise. 
        """
        message = preprocessed_input
        class EmotionExtractor(BaseModel):
            ContainsAnger: bool = Field(default=False)
            ContainsDisgust: bool = Field(default=False)
            ContainsFear: bool = Field(default=False)
            ContainsHappiness: bool = Field(default=False)
            ContainsSadness: bool = Field(default=False)
            ContainsSurprise: bool = Field(default=False)
    
        json_class = EmotionExtractor

        response = util.json_llm_call(system_prompt, message, json_class)

        result = [key[len('contains'):] for key in response.keys() if response[key]]

        return result

    def extract_titles(self, preprocessed_input):
        """Extract potential movie titles from a line of pre-processed text.

        Given an input text which has been pre-processed with preprocess(),
        this method should return a list of movie titles that are potentially
        in the text.

        - If there are no movie titles in the text, return an empty list.
        - If there is exactly one movie title in the text, return a list
        containing just that one movie title.
        - If there are multiple movie titles in the text, return a list
        of all movie titles you've extracted from the text.

        Example:
          potential_titles = chatbot.extract_titles(chatbot.preprocess(
                                            'I liked "The Notebook" a lot.'))
          print(potential_titles) // prints ["The Notebook"]

        :param preprocessed_input: a user-supplied line of text that has been
        pre-processed with preprocess()
        :returns: list of movie titles that are potentially in the text
        """
        matches = re.findall(r'"(.*?)"', preprocessed_input)
        
        return matches

    def find_movies_by_title(self, title):
        """ Given a movie title, return a list of indices of matching movies.

        - If no movies are found that match the given title, return an empty
        list.
        - If multiple movies are found that match the given title, return a list
        containing all of the indices of these matching movies.
        - If exactly one movie is found that matches the given title, return a
        list
        that contains the index of that matching movie.

        Example:
          ids = chatbot.find_movies_by_title('Titanic')
          print(ids) // prints [1359, 2716]

        :param title: a string containing a movie title
        :returns: a list of indices of matching movies
        """
        result_indexes = [] 
        title_year = re.search(r'\((\d{4})\)$', title)
        if title_year: 
            title = title[:-7]

        def add_matches(title, title_year, result_indexes):
            title = title.lower()
            with open('data/movies.txt', 'r') as file:
                file_lines = file.readlines()
                lines = [re.findall(r'%(.*?)%', line)[0][:-7]
                        for line in file_lines]
                fixed_lines = [re.sub(r'(.+?),\s+(The|A|An)\b', r'\2 \1', line)
                            for line in lines]
                years = [re.findall(r'\((\d{4})\)', line) for line in file_lines]
                movie_tuple_list = [tuple((fixed_lines[i], years[i]))
                                    for i in range(len(lines))]

                for i in range(len(movie_tuple_list)):
                    if title_year:
                        if title == movie_tuple_list[i][0].lower() and title_year.group(0)[1:-1] == movie_tuple_list[i][1][0]:
                            result_indexes.append(i)
                    else:
                        if title == movie_tuple_list[i][0].lower():
                            result_indexes.append(i)
        add_matches(title, title_year, result_indexes)
        if self.llm_enabled and len(result_indexes) == 0:
            system_prompt = """You are a bot that translates movie titles to English. Read the movie title given to you, """ +\
                """and translate the title to English. The translation should be the most common translation for that movie. If the title is already in English, then """ +\
                """you should respond with the original input and nothing else. You may encounter titles in various languages like German, Danish, Spanish, French, Italian, etc.""" +\
                """Only respond with just the translation in English and nothing more. Titles should not include notes in parantheses. For example, this is bad: 'Iron Man (original title is in Danish)'.\n\n"""
            stop = ["\n"]
            title = util.simple_llm_call(system_prompt, title, stop=stop).strip()
            title_year = re.search(r'\((\d{4})\)$', title)
            if title_year:
                title = title[:-7]
            add_matches(title, title_year, result_indexes)
        return result_indexes

    def extract_sentiment(self, preprocessed_input):
        """Extract a sentiment rating from a line of pre-processed text.

        You should return -1 if the sentiment of the text is negative, 0 if the
        sentiment of the text is neutral (no sentiment detected), or +1 if the
        sentiment of the text is positive.

        Example:
          sentiment = chatbot.extract_sentiment(chatbot.preprocess(
                                                    'I liked "The Titanic"'))
          print(sentiment) // prints 1

        :param preprocessed_input: a user-supplied line of text that has been
        pre-processed with preprocess()
        :returns: a numerical value for the sentiment of the text
        """
        stemmer = PorterStemmer()
        input = re.sub(re.compile(r'"([^"]*)"'), r'MOVIE', preprocessed_input)
        sentiment_dict = {}

        with open('data/sentiment.txt', 'r') as file:
            file_lines = file.readlines()
            lines = [line.strip() for line in file_lines]
            sentiment_tuples = [tuple(line.split(',')) for line in lines]
            for sentiment_tuple in sentiment_tuples: 
                sentiment_dict[stemmer.stem(sentiment_tuple[0])] = sentiment_tuple[1]
        
        negation_words = ['not', 'no', 'none', 'never', 'neither', 'nor', "didn't", "nothing", "nowhere", "without"]
        input_tokens = input.split()
        raw_score = 0 
        is_negative = False 
        for token in input_tokens:
            cleaned_token = re.sub(r'[,.;?!]', '', token.lower())
            stemmed_token = stemmer.stem(cleaned_token)
            if cleaned_token in negation_words: 
                is_negative = True
                
            if is_negative and re.search(r'[;,]', token):
                is_negative = False
            
            if stemmed_token in sentiment_dict.keys():
                token_score = 1 if sentiment_dict[stemmed_token] == 'pos' else -1
                negation_score = 1 if not is_negative else -1 
                raw_score += token_score * negation_score
        
        if raw_score > 0: 
            return 1 
        elif raw_score < 0: 
            return -1
        else: 
            return 0

    ############################################################################
    # 3. Movie Recommendation helper functions                                 #
    ############################################################################

    @staticmethod
    def binarize(ratings, threshold=2.5):
        """Return a binarized version of the given matrix.

        To binarize a matrix, replace all entries above the threshold with 1.
        and replace all entries at or below the threshold with a -1.

        Entries whose values are 0 represent null values and should remain at 0.

        Note that this method is intentionally made static, as you shouldn't use
        any attributes of Chatbot like self.ratings in this method.

        :param ratings: a (num_movies x num_users) matrix of user ratings, from
         0.5 to 5.0
        :param threshold: Numerical rating above which ratings are considered
        positive

        :returns: a binarized version of the movie-rating matrix
        """
        ########################################################################
        # TODO: Binarize the supplied ratings matrix.                          #
        #                                                                      #
        # WARNING: Do not use self.ratings directly in this function.          #
        ########################################################################

        # The starter code returns a new matrix shaped like ratings but full of
        # zeros.
        binarized_ratings = np.zeros_like(ratings)
        binarized_ratings[ratings > threshold] = 1
        binarized_ratings[(ratings <= threshold) & (ratings > 0)] = -1
        ########################################################################
        #                        END OF YOUR CODE                              #
        ########################################################################
        return binarized_ratings

    def similarity(self, u, v):
        """Calculate the cosine similarity between two vectors.

        You may assume that the two arguments have the same shape.

        :param u: one vector, as a 1D numpy array
        :param v: another vector, as a 1D numpy array

        :returns: the cosine similarity between the two vectors
        """
        ########################################################################
        # TODO: Compute cosine similarity between the two vectors.             #
        ########################################################################
        cosine_sim = np.dot(u,v)
        if(cosine_sim == 0):
            return 0
        cosine_sim /= (np.sqrt(np.dot(u,u)) * np.sqrt(np.dot(v,v)))
        ########################################################################
        #                          END OF YOUR CODE                            #
        ########################################################################
        return cosine_sim

    def recommend(self, user_ratings, ratings_matrix, k=10, llm_enabled=False):
        """Generate a list of indices of movies to recommend using collaborative
         filtering.

        You should return a collection of `k` indices of movies recommendations.

        As a precondition, user_ratings and ratings_matrix are both binarized.

        Remember to exclude movies the user has already rated!

        Please do not use self.ratings directly in this method.

        :param user_ratings: a binarized 1D numpy array of the user's movie
            ratings
        :param ratings_matrix: a binarized 2D numpy matrix of all ratings, where
          `ratings_matrix[i, j]` is the rating for movie i by user j
        :param k: the number of recommendations to generate
        :param llm_enabled: whether the chatbot is in llm programming mode

        :returns: a list of k movie indices corresponding to movies in
        ratings_matrix, in descending order of recommendation.
        """

        ########################################################################
        # TODO: Implement a recommendation function that takes a vector        #
        # user_ratings and matrix ratings_matrix and outputs a list of movies  #
        # recommended by the chatbot.                                          #
        #                                                                      #
        # WARNING: Do not use the self.ratings matrix directly in this         #
        # function.                                                            #
        #                                                                      #
        # For GUS mode, you should use item-item collaborative filtering with  #
        # cosine similarity, no mean-centering, and no normalization of        #
        # scores.                                                              #
        ########################################################################     
        num_items = len(user_ratings)
        recommendations = np.zeros(num_items)
        for i in range(num_items):
            if user_ratings[i] == 0:
                #If we have not rated this movie, don't use it
                continue
            for j in range(num_items): 
                if user_ratings[j] != 0:
                    # if we have rated this movie please don't recommend it
                    continue
                recommendations[j] += user_ratings[i] *self.similarity(ratings_matrix[i], ratings_matrix[j]) 
        ########################################################################
        #                        END OF YOUR CODE                              #
        ########################################################################
        return list(np.argsort(recommendations)[-k:][::-1])
    ############################################################################
    # 4. Debug info                                                            #
    ############################################################################

    def debug(self, line):
        """
        Return debug information as a string for the line string from the REPL

        NOTE: Pass the debug information that you may think is important for
        your evaluators.
        """
        debug_info = 'debug info'
        return debug_info

    ############################################################################
    # 5. Write a description for your chatbot here!                            #
    ############################################################################
    def intro(self):
        """Return a string to use as your chatbot's description for the user.

        Consider adding to this description any information about what your
        chatbot can do and how the user can interact with it.

        NOTE: This string will not be shown to the LLM in llm mode, this is just for the user
        """

        return """
        Your task is to implement the chatbot as detailed in the PA7
        instructions.
        Remember: in the GUS mode, movie names will come in quotation marks
        and expressions of sentiment will be simple!
        Explore movie recommendations with Archie, an elderly British movie enthusiast - it'll be a great time, innit?
        """


if __name__ == '__main__':
    print('To run your chatbot in an interactive loop from the command line, '
          'run:')
    print('    python3 repl.py')